@@ -1,7 +1,6 @@
 #![forbid(unsafe_code)]
 #![deny(
     warnings,
-    missing_docs,
     trivial_casts,
     trivial_numeric_casts,
     unused_import_braces,
@@ -10,7 +9,7 @@
     nonstandard_style
 )]
 #![doc(
-    html_root_url = "https://docs.rs/tendermint-light-client/0.21.0",
+    html_root_url = "https://docs.rs/tendermint-light-client/0.22.0",
     html_logo_url = "https://raw.githubusercontent.com/informalsystems/tendermint-rs/master/img/logo-tendermint-rs_3961x4001.png"
 )]
 #![cfg_attr(docsrs, feature(doc_cfg))]
@@ -34,8 +33,5 @@ pub mod types;
 
 pub(crate) mod utils;
 
-#[doc(hidden)]
-mod macros;
-
 #[doc(hidden)]
 pub mod tests;
